import gc
import warnings
from typing import List, Optional

import torch
import xarray as xr
from marshmallow import missing

from bioimageio.core.resource_io import nodes
from ._model_adapter import ModelAdapter


class PytorchModelAdapter(ModelAdapter):
    def _load(self, *, devices: Optional[List[str]] = None):
        self._model = self.get_nn_instance(self.bioimageio_model)

        if devices is None:
            self._devices = ["cuda" if torch.cuda.is_available() else "cpu"]
        else:
            self._devices = [torch.device(d) for d in devices]

        if len(self._devices) > 1:
            warnings.warn("Multiple devices for single pytorch model not yet implemented")

        self._model.to(self._devices[0])

        assert isinstance(self._model, torch.nn.Module)
        weights = self.bioimageio_model.weights.get("pytorch_state_dict")
        if weights is not None and weights.source:
            state = torch.load(weights.source, map_location=self._devices[0])
            self._model.load_state_dict(state)

        self._model.eval()
        self._internal_output_axes = [tuple(out.axes) for out in self.bioimageio_model.outputs]

    def _forward(self, *input_tensors: xr.DataArray) -> List[xr.DataArray]:
        with torch.no_grad():
            tensors = [torch.from_numpy(ipt.data) for ipt in input_tensors]
            tensors = [t.to(self._devices[0]) for t in tensors]
            result = self._model(*tensors)
            if not isinstance(result, (tuple, list)):
                result = [result]

            result = [r.detach().cpu().numpy() if isinstance(r, torch.Tensor) else r for r in result]

        return [xr.DataArray(r, dims=axes) for r, axes in zip(result, self._internal_output_axes)]

    def _unload(self) -> None:
        self._devices = None
        del self._model
        gc.collect()  # deallocate memory
        torch.cuda.empty_cache()  # release reserved memory

    @staticmethod
    def get_nn_instance(model_node: nodes.Model, **kwargs):
        weight_spec = model_node.weights.get("pytorch_state_dict")
        assert weight_spec is not None
        assert isinstance(weight_spec.architecture, nodes.ImportedSource)
        model_kwargs = weight_spec.kwargs
        joined_kwargs = {} if model_kwargs is missing else dict(model_kwargs)
        joined_kwargs.update(kwargs)
        return weight_spec.architecture(**joined_kwargs)



import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import ZipFile

from marshmallow import missing

from bioimageio.spec.model import raw_nodes


def test_export_package(any_onnx_model):
    from bioimageio.core import export_resource_package, load_raw_resource_description

    package_path = export_resource_package(any_onnx_model, weights_priority_order=["onnx"])
    assert isinstance(package_path, Path), package_path
    assert package_path.exists(), package_path

    raw_model = load_raw_resource_description(package_path)
    assert isinstance(raw_model, raw_nodes.Model)


def test_package_with_folder(unet2d_nuclei_broad_model):
    from bioimageio.core import export_resource_package, load_raw_resource_description

    with TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)

        # extract package (to not cache to BIOIMAGEIO_CACHE)
        package_folder = tmp_dir / "package"
        with ZipFile(unet2d_nuclei_broad_model) as zf:
            zf.extractall(package_folder)

        # load package
        model = load_raw_resource_description(package_folder / "rdf.yaml")
        assert isinstance(model, raw_nodes.Model)

        # alter package to have its documentation in a nested folder
        doc = model.documentation
        assert doc is not missing
        doc = doc.relative_to(model.root_path)
        assert not doc.is_absolute()
        new_doc = Path("nested") / "folder" / doc
        (package_folder / new_doc).parent.mkdir(parents=True)
        shutil.move(package_folder / doc, package_folder / new_doc)
        model.documentation = new_doc

        # export altered package
        altered_package = tmp_dir / "altered_package.zip"
        altered_package = export_resource_package(model, output_path=altered_package, weights_priority_order=["onnx"])

        # extract altered package (to not cache to BIOIMAGEIO_CACHE)
        altered_package_folder = tmp_dir / "altered_package"
        with ZipFile(altered_package) as zf:
            zf.extractall(altered_package_folder)

        # load altered package
        reloaded_model = load_raw_resource_description(altered_package_folder / "rdf.yaml")
        assert isinstance(reloaded_model, raw_nodes.Model)
        assert reloaded_model.documentation.as_posix().endswith(new_doc.as_posix())
        assert reloaded_model.documentation.exists()



import numpy as np
import pytest
import xarray as xr

from bioimageio.core.prediction_pipeline._measure_groups import compute_measures


def test_binarize():
    from bioimageio.core.prediction_pipeline._processing import Binarize

    shape = (3, 32, 32)
    axes = ("c", "y", "x")
    np_data = np.random.rand(*shape)
    data = xr.DataArray(np_data, dims=axes)

    threshold = 0.5
    exp = xr.DataArray(np_data > threshold, dims=axes)

    binarize = Binarize("data_name", threshold=threshold)
    res = binarize(data)
    xr.testing.assert_allclose(res, exp)


@pytest.mark.parametrize("axes", [None, tuple("cy"), tuple("cyx"), tuple("x")])
def test_scale_mean_variance(axes):
    from bioimageio.core.prediction_pipeline._processing import ScaleMeanVariance

    shape = (3, 32, 46)
    ipt_axes = ("c", "y", "x")
    np_data = np.random.rand(*shape)
    ipt_data = xr.DataArray(np_data, dims=ipt_axes)
    ref_data = xr.DataArray((np_data * 2) + 3, dims=ipt_axes)

    scale_mean_variance = ScaleMeanVariance("data_name", reference_tensor="ref_name", axes=axes)
    required = scale_mean_variance.get_required_measures()
    computed = compute_measures(required, sample={"data_name": ipt_data, "ref_name": ref_data})
    scale_mean_variance.set_computed_measures(computed)

    res = scale_mean_variance(ipt_data)
    xr.testing.assert_allclose(res, ref_data)


@pytest.mark.parametrize("axes", [None, tuple("cy"), tuple("y"), tuple("yx")])
def test_scale_mean_variance_per_channel(axes):
    from bioimageio.core.prediction_pipeline._processing import ScaleMeanVariance

    shape = (3, 32, 46)
    ipt_axes = ("c", "y", "x")
    np_data = np.random.rand(*shape)
    ipt_data = xr.DataArray(np_data, dims=ipt_axes)

    # set different mean, std per channel
    np_ref_data = np.stack([d * i + i for i, d in enumerate(np_data, start=2)])
    print(np_ref_data.shape)
    ref_data = xr.DataArray(np_ref_data, dims=ipt_axes)

    scale_mean_variance = ScaleMeanVariance("data_name", reference_tensor="ref_name", axes=axes)
    required = scale_mean_variance.get_required_measures()
    computed = compute_measures(required, sample={"data_name": ipt_data, "ref_name": ref_data})
    scale_mean_variance.set_computed_measures(computed)

    res = scale_mean_variance(ipt_data)

    if axes is not None and "c" not in axes:
        # mean,std per channel should match exactly
        xr.testing.assert_allclose(res, ref_data)
    else:
        # mean,std across channels should not match
        with pytest.raises(AssertionError):
            xr.testing.assert_allclose(res, ref_data)



import numpy as np


def test_transform_input_image():
    from bioimageio.core.image_helper import transform_input_image

    ax_list = ["yx", "xy", "cyx", "yxc", "bczyx", "xyz", "xyzc", "bzyxc"]
    im = np.random.rand(256, 256)
    for axes in ax_list:
        inp = transform_input_image(im, axes)
        assert inp.ndim == len(axes)

    ax_list = ["zyx", "cyx", "yxc", "bczyx", "xyz", "xyzc", "bzyxc"]
    vol = np.random.rand(64, 64, 64)
    for axes in ax_list:
        inp = transform_input_image(vol, axes)
        assert inp.ndim == len(axes)


def test_transform_output_tensor():
    from bioimageio.core.image_helper import transform_output_tensor

    tensor = np.random.rand(1, 3, 64, 64, 64)
    tensor_axes = "bczyx"

    out_ax_list = ["bczyx", "cyx", "xyc", "byxc", "zyx", "xyz"]
    for out_axes in out_ax_list:
        out = transform_output_tensor(tensor, tensor_axes, out_axes)
        assert out.ndim == len(out_axes)



import enum
import json
import os
import sys
import warnings
from glob import glob

from pathlib import Path
from pprint import pformat, pprint
from typing import List, Optional

import typer

from bioimageio.core import __version__, prediction, commands, resource_tests, load_raw_resource_description
from bioimageio.core.common import TestSummary
from bioimageio.core.prediction_pipeline import get_weight_formats
from bioimageio.spec.__main__ import app, help_version as help_version_spec
from bioimageio.spec.model.raw_nodes import WeightsFormat

try:
    from typing import get_args
except ImportError:
    from typing_extensions import get_args  # type: ignore

try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        from bioimageio.core.weight_converter import torch as torch_converter
except ImportError:
    torch_converter = None

try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        from bioimageio.core.weight_converter import keras as keras_converter
except ImportError:
    keras_converter = None


# extend help/version string by core version
help_version_core = f"bioimageio.core {__version__}"
help_version = f"{help_version_spec}\n{help_version_core}"
# prevent rewrapping with \b\n: https://click.palletsprojects.com/en/7.x/documentation/#preventing-rewrapping
app.info.help = "\b\n" + help_version


@app.callback()
def callback():
    typer.echo(help_version)


@app.command()
def package(
    rdf_source: str = typer.Argument(..., help="RDF source as relative file path or URI"),
    path: Path = typer.Argument(Path() / "{src_name}-package.zip", help="Save package as"),
    weights_priority_order: Optional[List[str]] = typer.Option(
        None,
        "--weights-priority-order",
        "-wpo",
        help="For model packages only. "
        "If given only the first weights matching the given weight formats are included. "
        "Defaults to include all weights present in source.",
        show_default=False,
    ),
    verbose: bool = typer.Option(False, help="show traceback of exceptions"),
):
    # typer bug: typer returns empty tuple instead of None if weights_order_priority is not given
    weights_priority_order = weights_priority_order or None

    ret_code = commands.package(
        rdf_source=rdf_source, path=path, weights_priority_order=weights_priority_order, verbose=verbose
    )
    sys.exit(ret_code)


package.__doc__ = commands.package.__doc__


# if we want to use something like "choice" for the weight formats, we need to use an enum, see:
# https://github.com/tiangolo/typer/issues/182
WeightFormatEnum = enum.Enum("WeightFormatEnum", {wf: wf for wf in get_args(WeightsFormat)})
# Enum with in values does not work with click.Choice: https://github.com/pallets/click/issues/784
# so a simple Enum with auto int values is not an option:
# WeightFormatEnum = enum.Enum("WeightFormatEnum", get_args(WeightsFormat))


def _log_test_summaries(summaries: List[TestSummary], msg: str):
    # todo: improve logging of multiple test summaries
    ret_code = 0
    for summary in summaries:
        print(f"{summary['name']}: {summary['status']}")
        if summary["status"] != "passed":
            s = {
                k: v
                for k, v in summary.items()
                if k not in ("name", "status", "bioimageio_spec_version", "bioimageio_core_version")
            }
            tb = s.pop("traceback")
            if tb:
                print("traceback:")
                print("".join(tb))

            def show_part(part, show):
                if show:
                    line = f"{part}: "
                    print(line + pformat(show, width=min(80, 120 - len(line))).replace("\n", " " * len(line) + "\n"))

            for part in ["error", "warnings", "source_name"]:
                show_part(part, s.pop(part, None))

            for part in sorted(s.keys()):
                show_part(part, s[part])

            ret_code = 1

    if ret_code:
        result = "FAILED!"
        icon = "❌"
    else:
        result = "passed."
        icon = "✔️"

    print(msg.format(icon=icon, result=result))
    return ret_code


@app.command()
def test_model(
    model_rdf: str = typer.Argument(
        ..., help="Path or URL to the model resource description file (rdf.yaml) or zipped model."
    ),
    weight_format: Optional[WeightFormatEnum] = typer.Option(None, help="The weight format to use."),
    devices: Optional[List[str]] = typer.Option(None, help="Devices for running the model."),
    decimal: int = typer.Option(4, help="The test precision."),
):
    # this is a weird typer bug: default devices are empty tuple although they should be None
    devices = devices or None

    summaries = resource_tests.test_model(
        model_rdf,
        weight_format=None if weight_format is None else weight_format.value,
        devices=devices,
        decimal=decimal,
    )
    print(f"\ntesting model {model_rdf}...")
    ret_code = _log_test_summaries(summaries, f"\n{{icon}} Model {model_rdf} {{result}}")
    sys.exit(ret_code)


test_model.__doc__ = resource_tests.test_model.__doc__


@app.command()
def test_resource(
    rdf: str = typer.Argument(
        ..., help="Path or URL to the resource description file (rdf.yaml) or zipped resource package."
    ),
    weight_format: Optional[WeightFormatEnum] = typer.Option(None, help="(for model only) The weight format to use."),
    devices: Optional[List[str]] = typer.Option(None, help="(for model only) Devices for running the model."),
    decimal: int = typer.Option(4, help="(for model only) The test precision."),
):
    # this is a weird typer bug: default devices are empty tuple although they should be None
    if len(devices) == 0:
        devices = None
    summaries = resource_tests.test_resource(
        rdf, weight_format=None if weight_format is None else weight_format.value, devices=devices, decimal=decimal
    )
    print(f"\ntesting {rdf}...")
    ret_code = _log_test_summaries(summaries, f"{{icon}} Resource test for {rdf} has {{result}}")
    sys.exit(ret_code)


test_resource.__doc__ = resource_tests.test_resource.__doc__


@app.command()
def predict_image(
    model_rdf: Path = typer.Argument(
        ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
    ),
    inputs: List[Path] = typer.Option(..., help="Path(s) to the model input(s)."),
    outputs: List[Path] = typer.Option(..., help="Path(s) for saveing the model output(s)."),
    # NOTE: typer currently doesn't support union types, so we only support boolean here
    # padding: Optional[Union[str, bool]] = typer.Argument(
    #     None, help="Padding to apply in each dimension passed as json encoded string."
    # ),
    # tiling: Optional[Union[str, bool]] = typer.Argument(
    #     None, help="Padding to apply in each dimension passed as json encoded string."
    # ),
    padding: Optional[bool] = typer.Option(None, help="Whether to pad the image to a size suited for the model."),
    tiling: Optional[bool] = typer.Option(None, help="Whether to run prediction in tiling mode."),
    weight_format: Optional[WeightFormatEnum] = typer.Option(None, help="The weight format to use."),
    devices: Optional[List[str]] = typer.Option(None, help="Devices for running the model."),
):

    if isinstance(padding, str):
        padding = json.loads(padding.replace("'", '"'))
        assert isinstance(padding, dict)
    if isinstance(tiling, str):
        tiling = json.loads(tiling.replace("'", '"'))
        assert isinstance(tiling, dict)

    # this is a weird typer bug: default devices are empty tuple although they should be None
    if len(devices) == 0:
        devices = None
    prediction.predict_image(
        model_rdf, inputs, outputs, padding, tiling, None if weight_format is None else weight_format.value, devices
    )


predict_image.__doc__ = prediction.predict_image.__doc__


@app.command()
def predict_images(
    model_rdf: Path = typer.Argument(
        ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
    ),
    input_pattern: str = typer.Argument(..., help="Glob pattern for the input images."),
    output_folder: str = typer.Argument(..., help="Folder to save the outputs."),
    output_extension: Optional[str] = typer.Argument(None, help="Optional output extension."),
    # NOTE: typer currently doesn't support union types, so we only support boolean here
    # padding: Optional[Union[str, bool]] = typer.Argument(
    #     None, help="Padding to apply in each dimension passed as json encoded string."
    # ),
    # tiling: Optional[Union[str, bool]] = typer.Argument(
    #     None, help="Padding to apply in each dimension passed as json encoded string."
    # ),
    padding: Optional[bool] = typer.Option(None, help="Whether to pad the image to a size suited for the model."),
    tiling: Optional[bool] = typer.Option(None, help="Whether to run prediction in tiling mode."),
    weight_format: Optional[WeightFormatEnum] = typer.Option(None, help="The weight format to use."),
    devices: Optional[List[str]] = typer.Option(None, help="Devices for running the model."),
):
    input_files = glob(input_pattern)
    input_names = [os.path.split(infile)[1] for infile in input_files]
    output_files = [os.path.join(output_folder, fname) for fname in input_names]
    if output_extension is not None:
        output_files = [f"{os.path.splitext(outfile)[0]}{output_extension}" for outfile in output_files]

    if isinstance(padding, str):
        padding = json.loads(padding.replace("'", '"'))
        assert isinstance(padding, dict)
    if isinstance(tiling, str):
        tiling = json.loads(tiling.replace("'", '"'))
        assert isinstance(tiling, dict)

    # this is a weird typer bug: default devices are empty tuple although they should be None
    if len(devices) == 0:
        devices = None
    prediction.predict_images(
        model_rdf,
        input_files,
        output_files,
        padding=padding,
        tiling=tiling,
        weight_format=None if weight_format is None else weight_format.value,
        devices=devices,
        verbose=True,
    )


predict_images.__doc__ = prediction.predict_images.__doc__


if torch_converter is not None:

    @app.command()
    def convert_torch_weights_to_onnx(
        model_rdf: Path = typer.Argument(
            ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
        ),
        output_path: Path = typer.Argument(..., help="Where to save the onnx weights."),
        opset_version: Optional[int] = typer.Argument(12, help="Onnx opset version."),
        use_tracing: bool = typer.Option(True, help="Whether to use torch.jit tracing or scripting."),
        verbose: bool = typer.Option(True, help="Verbosity"),
    ):
        ret_code = torch_converter.convert_weights_to_onnx(model_rdf, output_path, opset_version, use_tracing, verbose)
        sys.exit(ret_code)

    convert_torch_weights_to_onnx.__doc__ = torch_converter.convert_weights_to_onnx.__doc__

    @app.command()
    def convert_torch_weights_to_torchscript(
        model_rdf: Path = typer.Argument(
            ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
        ),
        output_path: Path = typer.Argument(..., help="Where to save the torchscript weights."),
        use_tracing: bool = typer.Option(True, help="Whether to use torch.jit tracing or scripting."),
    ):
        ret_code = torch_converter.convert_weights_to_torchscript(model_rdf, output_path, use_tracing)
        sys.exit(ret_code)

    convert_torch_weights_to_torchscript.__doc__ = torch_converter.convert_weights_to_torchscript.__doc__


if keras_converter is not None:

    @app.command()
    def convert_keras_weights_to_tensorflow(
        model_rdf: Path = typer.Argument(
            ..., help="Path to the model resource description file (rdf.yaml) or zipped model."
        ),
        output_path: Path = typer.Argument(..., help="Where to save the tensorflow weights."),
    ):
        ret_code = keras_converter.convert_weights_to_tensorflow_saved_model_bundle(model_rdf, output_path)
        sys.exit(ret_code)

    convert_keras_weights_to_tensorflow.__doc__ = (
        keras_converter.convert_weights_to_tensorflow_saved_model_bundle.__doc__
    )


if __name__ == "__main__":
    app()



import dataclasses
import importlib.util
import os
import pathlib
import sys
import typing
from types import ModuleType

from bioimageio.spec.shared import raw_nodes, resolve_source, source_available
from bioimageio.spec.shared.node_transformer import (
    GenericRawNode,
    GenericResolvedNode,
    NodeTransformer,
    NodeVisitor,
    UriNodeTransformer,
)
from . import nodes

GenericNode = typing.Union[GenericRawNode, GenericResolvedNode]


def iter_fields(node: GenericNode):
    for field in dataclasses.fields(node):
        yield field.name, getattr(node, field.name)


class SourceNodeChecker(NodeVisitor):
    """raises FileNotFoundError for unavailable URIs and paths"""

    def __init__(self, *, root_path: os.PathLike):
        self.root_path = root_path if isinstance(root_path, raw_nodes.URI) else pathlib.Path(root_path).resolve()

    def _visit_source(self, leaf: typing.Union[pathlib.Path, raw_nodes.URI]):
        if not source_available(leaf, self.root_path):
            raise FileNotFoundError(leaf)

    def visit_URI(self, node: raw_nodes.URI):
        self._visit_source(node)

    def visit_PosixPath(self, leaf: pathlib.PosixPath):
        self._visit_source(leaf)

    def visit_WindowsPath(self, leaf: pathlib.WindowsPath):
        self._visit_source(leaf)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""

        if isinstance(node, raw_nodes.RawNode):
            for field, value in iter_fields(node):
                if field != "root_path":  # do not visit root_path, as it might be an incomplete (non-available) URL
                    self.visit(value)
        else:
            super().generic_visit(node)


class SourceNodeTransformer(NodeTransformer):
    """
    Imports all source callables
    note: Requires previous transformation by UriNodeTransformer
    """

    class TemporaryInsertionIntoPythonPath:
        def __init__(self, path: str):
            self.path = path

        def __enter__(self):
            sys.path.insert(0, self.path)

        def __exit__(self, exc_type, exc_value, traceback):
            sys.path.remove(self.path)

    def transform_LocalImportableModule(self, node: raw_nodes.LocalImportableModule) -> nodes.ImportedSource:
        with self.TemporaryInsertionIntoPythonPath(str(node.root_path)):
            module = importlib.import_module(node.module_name)

        return nodes.ImportedSource(factory=getattr(module, node.callable_name))

    @staticmethod
    def transform_ResolvedImportableSourceFile(node: raw_nodes.ResolvedImportableSourceFile) -> nodes.ImportedSource:
        module_path = resolve_source(node.source_file)
        module_name = f"module_from_source.{module_path.stem}"
        importlib_spec = importlib.util.spec_from_file_location(module_name, module_path)
        assert importlib_spec is not None
        dep = importlib.util.module_from_spec(importlib_spec)
        importlib_spec.loader.exec_module(dep)  # type: ignore  # todo: possible to use "loader.load_module"?
        return nodes.ImportedSource(factory=getattr(dep, node.callable_name))


class RawNodeTypeTransformer(NodeTransformer):
    def __init__(self, nodes_module: ModuleType):
        super().__init__()
        self.nodes = nodes_module

    def generic_transformer(self, node: GenericRawNode) -> GenericResolvedNode:
        if isinstance(node, raw_nodes.RawNode):
            resolved_data = {
                field.name: self.transform(getattr(node, field.name)) for field in dataclasses.fields(node)
            }
            resolved_node_type: typing.Type[GenericResolvedNode] = getattr(self.nodes, node.__class__.__name__)
            return resolved_node_type(**resolved_data)  # type: ignore
        else:
            return super().generic_transformer(node)


def all_sources_available(
    node: typing.Union[GenericNode, list, tuple, dict], root_path: os.PathLike = pathlib.Path()
) -> bool:
    try:
        SourceNodeChecker(root_path=root_path).visit(node)
    except FileNotFoundError:
        return False
    else:
        return True


def resolve_raw_node(
    raw_rd: GenericRawNode, nodes_module: typing.Any, uri_only_if_in_package: bool = True
) -> GenericResolvedNode:
    """resolve all uris and paths (that are included when packaging)"""
    rd = UriNodeTransformer(root_path=raw_rd.root_path, uri_only_if_in_package=uri_only_if_in_package).transform(raw_rd)
    rd = SourceNodeTransformer().transform(rd)
    rd = RawNodeTypeTransformer(nodes_module).transform(rd)
    return rd



import os
from copy import deepcopy
from typing import Dict, List, Optional, Sequence, Tuple, Union

import imageio
import numpy as np
from xarray import DataArray
from bioimageio.core.resource_io.nodes import InputTensor, OutputTensor


#
# helper functions to transform input images / output tensors to the required axes
#


def transform_input_image(image: np.ndarray, tensor_axes: str, image_axes: Optional[str] = None):
    """Transform input image into output tensor with desired axes.

    Args:
        image: the input image
        tensor_axes: the desired tensor axes
        input_axes: the axes of the input image (optional)
    """
    # if the image axes are not given deduce them from the required axes and image shape
    if image_axes is None:
        has_z_axis = "z" in tensor_axes
        ndim = image.ndim
        if ndim == 2:
            image_axes = "yx"
        elif ndim == 3:
            image_axes = "zyx" if has_z_axis else "cyx"
        elif ndim == 4:
            image_axes = "czyx"
        elif ndim == 5:
            image_axes = "bczyx"
        else:
            raise ValueError(f"Invalid number of image dimensions: {ndim}")
    tensor = DataArray(image, dims=tuple(image_axes))
    # expand the missing image axes
    missing_axes = tuple(set(tensor_axes) - set(image_axes))
    tensor = tensor.expand_dims(dim=missing_axes)
    # transpose to the correct axis order
    tensor = tensor.transpose(*tuple(tensor_axes))
    # return numpy array
    return tensor.values


def _drop_axis_default(axis_name, axis_len):
    # spatial axes: drop at middle coordnate
    # other axes (channel or batch): drop at 0 coordinate
    return axis_len // 2 if axis_name in "zyx" else 0


def transform_output_tensor(tensor: np.ndarray, tensor_axes: str, output_axes: str, drop_function=_drop_axis_default):
    """Transform output tensor into image with desired axes.

    Args:
        tensor: the output tensor
        tensor_axes: bioimageio model spec
        output_axes: the desired output axes
        drop_function: function that determines how to drop unwanted axes
    """
    if len(tensor_axes) != tensor.ndim:
        raise ValueError(f"Number of axes {len(tensor_axes)} and dimension of tensor {tensor.ndim} don't match")
    shape = {ax_name: sh for ax_name, sh in zip(tensor_axes, tensor.shape)}
    output = DataArray(tensor, dims=tuple(tensor_axes))
    # drop unwanted axes
    drop_axis_names = tuple(set(tensor_axes) - set(output_axes))
    drop_axes = {ax_name: drop_function(ax_name, shape[ax_name]) for ax_name in drop_axis_names}
    output = output[drop_axes]
    # transpose to the desired axis order
    output = output.transpose(*tuple(output_axes))
    # return numpy array
    return output.values


def to_channel_last(image):
    chan_id = image.dims.index("c")
    if chan_id != image.ndim - 1:
        target_axes = tuple(ax for ax in image.dims if ax != "c") + ("c",)
        image = image.transpose(*target_axes)
    return image


#
# helper functions for loading and saving images
#


def load_image(in_path, axes: Sequence[str]) -> DataArray:
    ext = os.path.splitext(in_path)[1]
    if ext == ".npy":
        im = np.load(in_path)
    else:
        is_volume = "z" in axes
        im = imageio.volread(in_path) if is_volume else imageio.imread(in_path)
        im = transform_input_image(im, axes)
    return DataArray(im, dims=axes)


def load_tensors(sources, tensor_specs: List[Union[InputTensor, OutputTensor]]) -> List[DataArray]:
    return [load_image(s, sspec.axes) for s, sspec in zip(sources, tensor_specs)]


def save_image(out_path, image):
    ext = os.path.splitext(out_path)[1]
    if ext == ".npy":
        np.save(out_path, image)
    else:
        is_volume = "z" in image.dims

        # squeeze batch or channel axes if they are singletons
        squeeze = {ax: 0 if (ax in "bc" and sh == 1) else slice(None) for ax, sh in zip(image.dims, image.shape)}
        image = image[squeeze]

        if "b" in image.dims:
            raise RuntimeError(f"Cannot save prediction with batchsize > 1 as {ext}-file")
        if "c" in image.dims:  # image formats need channel last
            image = to_channel_last(image)

        save_function = imageio.volsave if is_volume else imageio.imsave
        # most image formats only support channel dimensions of 1, 3 or 4;
        # if not we need to save the channels separately
        ndim = 3 if is_volume else 2
        save_as_single_image = image.ndim == ndim or (image.shape[-1] in (3, 4))

        if save_as_single_image:
            save_function(out_path, image)
        else:
            out_prefix, ext = os.path.splitext(out_path)
            for c in range(image.shape[-1]):
                chan_out_path = f"{out_prefix}-c{c}{ext}"
                save_function(chan_out_path, image[..., c])


#
# helper function for padding
#


def pad(image, axes: Sequence[str], padding, pad_right=True) -> Tuple[np.ndarray, Dict[str, slice]]:
    assert image.ndim == len(axes), f"{image.ndim}, {len(axes)}"

    padding_ = deepcopy(padding)
    mode = padding_.pop("mode", "dynamic")
    assert mode in ("dynamic", "fixed")

    is_volume = "z" in axes
    if is_volume:
        assert len(padding_) == 3
    else:
        assert len(padding_) == 2

    if isinstance(pad_right, bool):
        pad_right = len(axes) * [pad_right]

    pad_width = []
    crop = {}
    for ax, dlen, pr in zip(axes, image.shape, pad_right):

        if ax in "zyx":
            pad_to = padding_[ax]

            if mode == "dynamic":
                r = dlen % pad_to
                pwidth = 0 if r == 0 else (pad_to - r)
            else:
                if pad_to < dlen:
                    msg = f"Padding for axis {ax} failed; pad shape {pad_to} is smaller than the image shape {dlen}."
                    raise RuntimeError(msg)
                pwidth = pad_to - dlen

            pad_width.append([0, pwidth] if pr else [pwidth, 0])
            crop[ax] = slice(0, dlen) if pr else slice(pwidth, None)
        else:
            pad_width.append([0, 0])
            crop[ax] = slice(None)

    image = np.pad(image, pad_width, mode="symmetric")
    return image, crop



"""Here pre- and postprocessing operations are implemented according to their definitions in bioimageio.spec:
see https://github.com/bioimage-io/spec-bioimage-io/blob/gh-pages/preprocessing_spec_latest.md
and https://github.com/bioimage-io/spec-bioimage-io/blob/gh-pages/postprocessing_spec_latest.md
"""
import numbers
from dataclasses import InitVar, dataclass, field, fields
from typing import List, Mapping, Optional, Sequence, Tuple, Type, Union

import numpy
import numpy as np
import xarray as xr

from bioimageio.core.statistical_measures import Mean, Measure, Percentile, Std
from bioimageio.spec.model.raw_nodes import PostprocessingName, PreprocessingName
from ._utils import ComputedMeasures, DatasetMode, FIXED, Mode, PER_DATASET, PER_SAMPLE, RequiredMeasures, SampleMode

try:
    from typing import Literal, get_args, TypedDict
except ImportError:
    from typing_extensions import Literal, get_args, TypedDict  # type: ignore


def _get_fixed(
    fixed: Union[float, Sequence[float]], tensor: xr.DataArray, axes: Optional[Sequence[str]]
) -> Union[float, xr.DataArray]:
    if axes is None:
        return fixed

    fixed_shape = tuple(s for d, s in tensor.sizes.items() if d not in axes)
    fixed_dims = tuple(d for d in tensor.dims if d not in axes)
    fixed = np.array(fixed).reshape(fixed_shape)
    return xr.DataArray(fixed, dims=fixed_dims)


TensorName = str

MISSING = "MISSING"


@dataclass
class Processing:
    """base class for all Pre- and Postprocessing transformations."""

    tensor_name: str
    # todo: in python>=3.10 we should use dataclasses.KW_ONLY instead of MISSING (see child classes) to make inheritance work properly
    computed_measures: ComputedMeasures = field(default_factory=dict)
    mode: Mode = FIXED

    def get_required_measures(self) -> RequiredMeasures:
        return {}

    def set_computed_measures(self, computed: ComputedMeasures):
        # check if computed contains all required measures
        for mode, req_per_mode in self.get_required_measures().items():
            for tn, req_per_tn in req_per_mode.items():
                comp_measures = computed.get(mode, {}).get(tn, {})
                for req_measure in req_per_tn:
                    if req_measure not in comp_measures:
                        raise ValueError(f"Missing required {req_measure} for {tn} {mode}.")

        self.computed_measures = computed

    def get_computed_measure(self, tensor_name: TensorName, measure: Measure, *, mode: Optional[Mode] = None):
        """helper to unpack self.computed_measures"""
        ret = self.computed_measures.get(mode or self.mode, {}).get(tensor_name, {}).get(measure)
        if ret is None:
            raise RuntimeError(f"Missing computed {measure} for {tensor_name} {mode}.")

        return ret

    def __call__(self, tensor: xr.DataArray) -> xr.DataArray:
        return self.apply(tensor)

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        """apply processing"""
        raise NotImplementedError

    def __post_init__(self):
        # validate common kwargs by their annotations
        for f in fields(self):
            # check MISSING
            if getattr(self, f.name) is MISSING:
                raise TypeError(f"missing required argument {f.name}")

            if f.name == "mode":
                # mode is always annotated as literals (or literals of literals)
                valid_modes = get_args(f.type)
                for inner in get_args(f.type):
                    valid_modes += get_args(inner)

                if self.mode not in valid_modes:
                    raise NotImplementedError(f"Unsupported mode {self.mode} for {self.__class__.__name__}")


#
# Pre- and Postprocessing implementations
#


@dataclass
class AssertDtype(Processing):
    """Helper Processing to assert dtype."""

    dtype: Union[str, Sequence[str]] = MISSING
    assert_with: Tuple[Type[numpy.dtype], ...] = field(init=False)

    def __post_init__(self):
        if isinstance(self.dtype, str):
            dtype = [self.dtype]
        else:
            dtype = self.dtype

        self.assert_with = tuple(type(numpy.dtype(dt)) for dt in dtype)

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        assert isinstance(tensor.dtype, self.assert_with)
        return tensor


@dataclass
class Binarize(Processing):
    """'output = tensor > threshold'."""

    threshold: float = MISSING  # make dataclass inheritance work for py<3.10 by using an explicit MISSING value.

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return tensor > self.threshold


@dataclass
class Clip(Processing):
    """Limit tensor values to [min, max]."""

    min: float = MISSING
    max: float = MISSING

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return tensor.clip(min=self.min, max=self.max)


@dataclass
class EnsureDtype(Processing):
    """Helper Processing to cast dtype if needed."""

    dtype: str = MISSING

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return tensor.astype(self.dtype)


@dataclass
class ScaleLinear(Processing):
    """Scale the tensor with a fixed multiplicative and additive factor."""

    gain: Union[float, Sequence[float]] = MISSING
    offset: Union[float, Sequence[float]] = MISSING
    axes: Optional[Sequence[str]] = None

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        scale_axes = tuple(ax for ax in tensor.dims if (ax not in self.axes and ax != "b"))
        if scale_axes:
            gain = xr.DataArray(np.atleast_1d(self.gain), dims=scale_axes)
            offset = xr.DataArray(np.atleast_1d(self.offset), dims=scale_axes)
        else:
            gain = self.gain
            offset = self.offset

        return tensor * gain + offset

    def __post_init__(self):
        super().__post_init__()
        if self.axes is None:
            assert isinstance(self.gain, (int, float))
            assert isinstance(self.offset, (int, float))


@dataclass
class ScaleMeanVariance(Processing):
    """Scale the tensor s.t. its mean and variance match a reference tensor."""

    mode: Literal[SampleMode, DatasetMode] = PER_SAMPLE
    reference_tensor: TensorName = MISSING
    axes: Optional[Sequence[str]] = None
    eps: float = 1e-6

    def get_required_measures(self) -> RequiredMeasures:
        axes = None if self.axes is None else tuple(self.axes)
        return {
            self.mode: {
                self.tensor_name: {Mean(axes=axes), Std(axes=axes)},
                self.reference_tensor: {Mean(axes=axes), Std(axes=axes)},
            }
        }

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        axes = None if self.axes is None else tuple(self.axes)
        assert self.mode in (PER_SAMPLE, PER_DATASET)
        mean = self.get_computed_measure(self.tensor_name, Mean(axes), mode=self.mode)
        std = self.get_computed_measure(self.tensor_name, Std(axes), mode=self.mode)
        ref_mean = self.get_computed_measure(self.reference_tensor, Mean(axes), mode=self.mode)
        ref_std = self.get_computed_measure(self.reference_tensor, Std(axes), mode=self.mode)

        return (tensor - mean) / (std + self.eps) * (ref_std + self.eps) + ref_mean


@dataclass
class ScaleRange(Processing):
    """Scale with percentiles."""

    mode: Literal[SampleMode, DatasetMode] = PER_SAMPLE
    axes: Optional[Sequence[str]] = None
    min_percentile: float = 0.0
    max_percentile: float = 100.0
    eps: float = 1e-6
    reference_tensor: Optional[TensorName] = None

    def get_required_measures(self) -> RequiredMeasures:
        axes = None if self.axes is None else tuple(self.axes)
        measures = {Percentile(self.min_percentile, axes=axes), Percentile(self.max_percentile, axes=axes)}
        return {self.mode: {self.reference_tensor or self.tensor_name: measures}}

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        ref_name = self.reference_tensor or self.tensor_name
        axes = None if self.axes is None else tuple(self.axes)
        v_lower = self.get_computed_measure(ref_name, Percentile(self.min_percentile, axes=axes))
        v_upper = self.get_computed_measure(ref_name, Percentile(self.max_percentile, axes=axes))

        return (tensor - v_lower) / (v_upper - v_lower + self.eps)

    def __post_init__(self):
        super().__post_init__()
        self.axes = None if self.axes is None else tuple(self.axes)  # make sure axes is Tuple[str] or None


@dataclass
class Sigmoid(Processing):
    """1 / (1 + e^(-tensor))."""

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        return 1.0 / (1.0 + np.exp(-tensor))


@dataclass
class ZeroMeanUnitVariance(Processing):
    """normalize to zero mean, unit variance."""

    mode: Mode = PER_SAMPLE
    mean: Optional[Union[float, Sequence[float]]] = None
    std: Optional[Union[float, Sequence[float]]] = None
    axes: Optional[Sequence[str]] = None
    eps: float = 1.0e-6

    def get_required_measures(self) -> RequiredMeasures:
        if self.mode == FIXED:
            return {}
        else:
            axes = None if self.axes is None else tuple(self.axes)
            return {self.mode: {self.tensor_name: {Mean(axes=axes), Std(axes=axes)}}}

    def apply(self, tensor: xr.DataArray) -> xr.DataArray:
        axes = None if self.axes is None else tuple(self.axes)
        if self.mode == FIXED:
            assert self.mean is not None and self.std is not None
            mean = _get_fixed(self.mean, tensor, axes)
            std = _get_fixed(self.std, tensor, axes)
        elif self.mode in (PER_SAMPLE, PER_DATASET):
            assert self.mean is None and self.std is None
            mean = self.get_computed_measure(self.tensor_name, Mean(axes), mode=self.mode)
            std = self.get_computed_measure(self.tensor_name, Std(axes), mode=self.mode)
        else:
            raise ValueError(self.mode)

        return (tensor - mean) / (std + self.eps)


_KnownProcessing = TypedDict(
    "_KnownProcessing",
    dict(pre=Mapping[PreprocessingName, Type[Processing]], post=Mapping[PostprocessingName, Type[Processing]]),
)

KNOWN_PROCESSING: _KnownProcessing = dict(
    pre={
        "binarize": Binarize,
        "clip": Clip,
        "scale_linear": ScaleLinear,
        "scale_range": ScaleRange,
        "sigmoid": Sigmoid,
        "zero_mean_unit_variance": ZeroMeanUnitVariance,
    },
    post={
        "binarize": Binarize,
        "clip": Clip,
        "scale_linear": ScaleLinear,
        "scale_mean_variance": ScaleMeanVariance,
        "scale_range": ScaleRange,
        "sigmoid": Sigmoid,
        "zero_mean_unit_variance": ZeroMeanUnitVariance,
    },
)



import torch
from bioimageio.core.prediction_pipeline._model_adapters._pytorch_model_adapter import PytorchModelAdapter


# additional convenience for pytorch state dict, eventually we want this in python-bioimageio too
# and for each weight format
def load_model(node):
    model = PytorchModelAdapter.get_nn_instance(node)
    state = torch.load(node.weights["pytorch_state_dict"].source, map_location="cpu")
    model.load_state_dict(state)
    model.eval()
    return model



import os
import subprocess
from typing import Sequence

import numpy as np
import pytest

from bioimageio.core import load_resource_description


def run_subprocess(commands: Sequence[str], **kwargs) -> subprocess.CompletedProcess:
    return subprocess.run(commands, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding="utf-8", **kwargs)


def test_validate_model(unet2d_nuclei_broad_model):
    ret = run_subprocess(["bioimageio", "validate", unet2d_nuclei_broad_model])
    assert ret.returncode == 0, ret.stdout


def test_cli_package(unet2d_nuclei_broad_model, tmp_path):
    out_path = tmp_path / "model.zip"
    ret = run_subprocess(["bioimageio", "package", unet2d_nuclei_broad_model, str(out_path)])
    assert ret.returncode == 0, ret.stdout
    assert out_path.exists()


def test_cli_package_wo_cache(unet2d_nuclei_broad_model):
    env = os.environ.copy()
    env["BIOIMAGEIO_USE_CACHE"] = "false"
    ret = run_subprocess(["bioimageio", "package", unet2d_nuclei_broad_model], env=env)
    assert ret.returncode == 0, ret.stdout


def test_cli_test_model(unet2d_nuclei_broad_model):
    ret = run_subprocess(["bioimageio", "test-model", unet2d_nuclei_broad_model])
    assert ret.returncode == 0, ret.stdout


def test_cli_test_model_fail(stardist_wrong_shape):
    ret = run_subprocess(["bioimageio", "test-model", stardist_wrong_shape])
    assert ret.returncode == 1


def test_cli_test_model_with_weight_format(unet2d_nuclei_broad_model):
    ret = run_subprocess(
        ["bioimageio", "test-model", unet2d_nuclei_broad_model, "--weight-format", "pytorch_state_dict"]
    )
    assert ret.returncode == 0, ret.stdout


def test_cli_test_resource(unet2d_nuclei_broad_model):
    ret = run_subprocess(["bioimageio", "test-resource", unet2d_nuclei_broad_model])
    assert ret.returncode == 0, ret.stdout


def test_cli_test_resource_with_weight_format(unet2d_nuclei_broad_model):
    ret = run_subprocess(
        ["bioimageio", "test-resource", unet2d_nuclei_broad_model, "--weight-format", "pytorch_state_dict"]
    )
    assert ret.returncode == 0, ret.stdout


def _test_cli_predict_image(model, tmp_path, extra_kwargs=None):
    spec = load_resource_description(model)
    in_path = spec.test_inputs[0]
    out_path = tmp_path.with_suffix(".npy")
    cmd = ["bioimageio", "predict-image", model, "--inputs", str(in_path), "--outputs", str(out_path)]
    if extra_kwargs is not None:
        cmd.extend(extra_kwargs)
    ret = run_subprocess(cmd)
    assert ret.returncode == 0, ret.stdout
    assert out_path.exists()


def test_cli_predict_image(unet2d_nuclei_broad_model, tmp_path):
    _test_cli_predict_image(unet2d_nuclei_broad_model, tmp_path)


def test_cli_predict_image_with_weight_format(unet2d_nuclei_broad_model, tmp_path):
    _test_cli_predict_image(unet2d_nuclei_broad_model, tmp_path, ["--weight-format", "pytorch_state_dict"])


def _test_cli_predict_images(model, tmp_path, extra_kwargs=None):
    n_images = 3
    shape = (1, 1, 128, 128)
    expected_shape = (1, 1, 128, 128)

    in_folder = tmp_path / "inputs"
    in_folder.mkdir()
    out_folder = tmp_path / "outputs"
    out_folder.mkdir()

    expected_outputs = []
    for i in range(n_images):
        path = in_folder / f"im-{i}.npy"
        im = np.random.randint(0, 255, size=shape).astype("uint8")
        np.save(path, im)
        expected_outputs.append(out_folder / f"im-{i}.npy")

    input_pattern = str(in_folder / "*.npy")
    cmd = ["bioimageio", "predict-images", model, input_pattern, str(out_folder)]
    if extra_kwargs is not None:
        cmd.extend(extra_kwargs)
    ret = run_subprocess(cmd)
    assert ret.returncode == 0, ret.stdout

    for out_path in expected_outputs:
        assert out_path.exists()
        assert np.load(out_path).shape == expected_shape


def test_cli_predict_images(unet2d_nuclei_broad_model, tmp_path):
    _test_cli_predict_images(unet2d_nuclei_broad_model, tmp_path)


def test_cli_predict_images_with_weight_format(unet2d_nuclei_broad_model, tmp_path):
    _test_cli_predict_images(unet2d_nuclei_broad_model, tmp_path, ["--weight-format", "pytorch_state_dict"])


def test_torch_to_torchscript(unet2d_nuclei_broad_model, tmp_path):
    out_path = tmp_path.with_suffix(".pt")
    ret = run_subprocess(
        ["bioimageio", "convert-torch-weights-to-torchscript", str(unet2d_nuclei_broad_model), str(out_path)]
    )
    assert ret.returncode == 0, ret.stdout
    assert out_path.exists()


@pytest.mark.skipif(pytest.skip_onnx, reason="requires torch and onnx")
def test_torch_to_onnx(unet2d_nuclei_broad_model, tmp_path):
    out_path = tmp_path.with_suffix(".onnx")
    ret = run_subprocess(["bioimageio", "convert-torch-weights-to-onnx", str(unet2d_nuclei_broad_model), str(out_path)])
    assert ret.returncode == 0, ret.stdout
    assert out_path.exists()


def test_keras_to_tf(unet2d_keras, tmp_path):
    out_path = tmp_path / "weights.zip"
    ret = run_subprocess(["bioimageio", "convert-keras-weights-to-tensorflow", str(unet2d_keras), str(out_path)])
    assert ret.returncode == 0, ret.stdout
    assert out_path.exists()



import warnings

from pathlib import Path
from typing import Union

import numpy as np
import torch
from numpy.testing import assert_array_almost_equal

import bioimageio.spec as spec
from bioimageio.core import load_resource_description
from .utils import load_model


def _check_predictions(model, scripted_model, model_spec, input_data):
    assert isinstance(input_data, list)

    def _check(input_):
        # get the expected output to validate the torchscript weights
        expected_outputs = model(*input_)
        if isinstance(expected_outputs, (torch.Tensor)):
            expected_outputs = [expected_outputs]
        expected_outputs = [out.numpy() for out in expected_outputs]

        outputs = scripted_model(*input_)
        if isinstance(outputs, (torch.Tensor)):
            outputs = [outputs]
        outputs = [out.numpy() for out in outputs]

        try:
            for exp, out in zip(expected_outputs, outputs):
                assert_array_almost_equal(exp, out, decimal=4)
            return 0
        except AssertionError as e:
            msg = f"The onnx weights were exported, but results before and after conversion do not agree:\n {str(e)}"
            warnings.warn(msg)
            return 1

    ret = _check(input_data)
    n_inputs = len(model_spec.inputs)
    # check has not passed or we have more tahn one input? then return immediately
    if ret == 1 or n_inputs > 1:
        return ret

    # do we have fixed input size or variable?
    # if variable, we need to check multiple sizes!
    shape_spec = model_spec.inputs[0].shape
    try:  # we have a variable shape
        min_shape = shape_spec.min
        step = shape_spec.step
    except AttributeError:  # we have fixed shape
        return ret

    half_step = [st // 2 for st in step]
    max_steps = 4
    step_factor = 1

    # check that input and output agree for decreasing input sizes
    while True:

        slice_ = tuple(slice(None) if st == 0 else slice(step_factor * st, -step_factor * st) for st in half_step)
        this_input = [inp[slice_] for inp in input_data]
        this_shape = this_input[0].shape
        if any(tsh < msh for tsh, msh in zip(this_shape, min_shape)):
            return ret

        ret = _check(this_input)
        if ret == 1:
            return ret
        step_factor += 1
        if step_factor > max_steps:
            return ret


def convert_weights_to_torchscript(
    model_spec: Union[str, Path, spec.model.raw_nodes.Model], output_path: Union[str, Path], use_tracing: bool = True
):
    """Convert model weights from format 'pytorch_state_dict' to 'torchscript'.

    Args:
        model_spec: location of the resource for the input bioimageio model
        output_path: where to save the torchscript weights
        use_tracing: whether to use tracing or scripting to export the torchscript format
    """
    if isinstance(model_spec, (str, Path)):
        model_spec = load_resource_description(Path(model_spec))

    with torch.no_grad():
        # load input and expected output data
        input_data = [np.load(inp).astype("float32") for inp in model_spec.test_inputs]
        input_data = [torch.from_numpy(inp) for inp in input_data]

        # instantiate model and get reference output
        model = load_model(model_spec)

        # make scripted model
        if use_tracing:
            scripted_model = torch.jit.trace(model, input_data)
        else:
            scripted_model = torch.jit.script(model)

        # check the scripted model
        ret = _check_predictions(model, scripted_model, model_spec, input_data)

    # save the torchscript model
    scripted_model.save(str(output_path))  # does not support Path, so need to cast to str
    return ret



from __future__ import annotations

import collections
import warnings
from collections import defaultdict
from itertools import product
from typing import DefaultDict, Dict, Hashable, Iterator, List, Mapping, Optional, Sequence, Set, Tuple, Type, Union

import numpy
import xarray as xr

from bioimageio.core.statistical_measures import Mean, Measure, Percentile, Std, Var
from ._utils import ComputedMeasures, PER_DATASET, PER_SAMPLE, RequiredMeasures, Sample, TensorName

try:
    from typing import Literal, TypedDict
except ImportError:
    from typing_extensions import Literal, TypedDict  # type: ignore

try:
    import crick
except ImportError:
    crick = None

MeasureValue = xr.DataArray


class SampleMeasureGroup:
    """group of measures for more efficient computation of multiple measures per sample"""

    def compute(self, sample: Sample) -> Dict[TensorName, Dict[Measure, MeasureValue]]:
        raise NotImplementedError


class DatasetMeasureGroup:
    """group of measures for more efficient computation of multiple measures per dataset"""

    def reset(self) -> None:
        """reset any accumulated intermediates"""
        raise NotImplementedError

    def update_with_sample(self, sample: Sample) -> None:
        """update intermediate representation with a data sample"""
        raise NotImplementedError

    def finalize(self) -> Dict[TensorName, Dict[Measure, MeasureValue]]:
        """compute statistics from intermediate representation"""
        raise NotImplementedError


MeasureGroups = TypedDict(
    "MeasureGroups", {PER_SAMPLE: Sequence[SampleMeasureGroup], PER_DATASET: Sequence[DatasetMeasureGroup]}
)


class DatasetMean(DatasetMeasureGroup):
    n: int
    mean: Optional[xr.DataArray]

    def __init__(self, tensor_name: TensorName, axes: Optional[Tuple[int]]):
        self.axes: Optional[Tuple[str]] = axes
        self.tensor_name = tensor_name
        self.reset()

    def reset(self):
        self.n = 0
        self.mean = None

    def update_with_sample(self, sample: Sample):
        tensor = sample[self.tensor_name].astype(numpy.float64, copy=False)
        mean_b = tensor.mean(dim=self.axes)
        assert mean_b.dtype == numpy.float64
        n_b = numpy.prod(tensor.shape) / numpy.prod(mean_b.shape)  # reduced voxel count
        if self.n == 0:
            assert self.mean is None
            self.n = n_b
            self.mean = mean_b
        else:
            n_a = self.n
            mean_a = self.mean
            self.n = n = n_a + n_b
            self.mean = (n_a * mean_a + n_b * mean_b) / n
            assert self.mean.dtype == numpy.float64

    def finalize(self) -> Dict[TensorName, Dict[Measure, MeasureValue]]:
        if self.n == 0:
            return {}
        else:
            return {self.tensor_name: {Mean(axes=self.axes): self.mean}}


class MeanVarStd(SampleMeasureGroup, DatasetMeasureGroup):
    n: int
    mean: Optional[xr.DataArray]
    m2: Optional[xr.DataArray]

    def __init__(self, tensor_name: TensorName, axes: Optional[Tuple[int]]):
        self.axes: Optional[Tuple[str]] = axes
        self.tensor_name = tensor_name
        self.reset()

    def compute(self, sample: Sample) -> Dict[TensorName, Dict[Measure, MeasureValue]]:
        tensor = sample[self.tensor_name]
        mean = tensor.mean(dim=self.axes)
        c = tensor - mean
        n = tensor.size if self.axes is None else numpy.prod([tensor.sizes[d] for d in self.axes])
        var = xr.dot(c, c, dims=self.axes) / n
        std = numpy.sqrt(var)
        return {self.tensor_name: {Mean(axes=self.axes): mean, Var(axes=self.axes): var, Std(axes=self.axes): std}}

    def reset(self):
        self.n = 0
        self.mean = None
        self.m2 = None

    def update_with_sample(self, sample: Sample):
        tensor = sample[self.tensor_name].astype(numpy.float64, copy=False)
        mean_b = tensor.mean(dim=self.axes)
        assert mean_b.dtype == numpy.float64
        n_b = numpy.prod(tensor.shape) / numpy.prod(mean_b.shape)  # reduced voxel count
        m2_b = ((tensor - mean_b) ** 2).sum(dim=self.axes)
        assert m2_b.dtype == numpy.float64
        if self.n == 0:
            assert self.mean is None
            assert self.m2 is None
            self.n = n_b
            self.mean = mean_b
            self.m2 = m2_b
        else:
            n_a = self.n
            mean_a = self.mean
            m2_a = self.m2
            self.n = n = n_a + n_b
            self.mean = (n_a * mean_a + n_b * mean_b) / n
            assert self.mean.dtype == numpy.float64
            d = mean_b - mean_a
            self.m2 = m2_a + m2_b + d ** 2 * n_a * n_b / n
            assert self.m2.dtype == numpy.float64

    def finalize(self) -> Dict[TensorName, Dict[Measure, MeasureValue]]:
        if self.n == 0:
            return {}
        else:
            var = self.m2 / self.n
            return {
                self.tensor_name: {
                    Mean(axes=self.axes): self.mean,
                    Var(axes=self.axes): var,
                    Std(axes=self.axes): numpy.sqrt(var),
                }
            }


class SamplePercentiles(SampleMeasureGroup):
    def __init__(self, tensor_name: TensorName, axes: Optional[Tuple[str]], ns: Sequence[float]):
        assert all(0 <= n <= 100 for n in ns)
        self.ns = ns
        self.qs = [n / 100 for n in ns]
        self.axes = axes
        self.tensor_name = tensor_name

    def compute(self, sample: Sample) -> Dict[TensorName, Dict[Measure, MeasureValue]]:
        tensor = sample[self.tensor_name]
        ps = tensor.quantile(self.qs, dim=self.axes)
        return {self.tensor_name: {Percentile(n=n, axes=self.axes): p for n, p in zip(self.ns, ps)}}


class MeanPercentiles(DatasetMeasureGroup):
    n: int
    estimates: Optional[xr.DataArray]

    def __init__(self, tensor_name: TensorName, axes: Optional[Tuple[str]], ns: Sequence[float]):
        assert all(0 <= n <= 100 for n in ns)
        self.ns = ns
        self.qs = [n / 100 for n in ns]
        self.axes = axes
        self.tensor_name = tensor_name
        self.reset()

    def reset(self):
        self.n = 0
        self.estimates = None

    def update_with_sample(self, sample: Sample):
        tensor = sample[self.tensor_name]
        sample_estimates = tensor.quantile(self.qs, dim=self.axes).astype(numpy.float64, copy=False)

        n = numpy.prod(tensor.shape) / numpy.prod(sample_estimates.shape[1:])  # reduced voxel count

        if self.n == 0:
            self.estimates = sample_estimates
        else:
            self.estimates = (self.n * self.estimates + n * sample_estimates) / (self.n + n)
            assert self.estimates.dtype == numpy.float64

        self.n += n

    def finalize(self) -> Dict[TensorName, Dict[Percentile, MeasureValue]]:
        if self.n == 0:
            return {}
        else:
            warnings.warn(f"Computed dataset percentiles naively by averaging percentiles of samples.")
            return {self.tensor_name: {Percentile(n=n, axes=self.axes): e for n, e in zip(self.ns, self.estimates)}}


class CrickPercentiles(DatasetMeasureGroup):
    digest: Optional[List["crick.TDigest"]]
    dims: Optional[Tuple[Hashable, ...]]
    indices: Optional[Iterator[Tuple[int, ...]]]
    shape: Optional[Tuple[int, ...]]

    def __init__(self, tensor_name: TensorName, axes: Optional[Tuple[str]], ns: Sequence[float]):
        assert all(0 <= n <= 100 for n in ns)
        assert axes is None or "_percentiles" not in axes
        warnings.warn(f"Computing dataset percentiles with experimental 'crick' library.")
        self.ns = ns
        self.qs = [n / 100 for n in ns]
        self.axes = axes
        self.tensor_name = tensor_name
        self.reset()

    def reset(self):
        self.digest = None
        self.dims = None
        self.indices = None
        self.shape = None

    def _initialize(self, tensor_sizes: Mapping[Hashable, int]):
        out_sizes = collections.OrderedDict(_percentiles=len(self.ns))
        if self.axes is not None:
            for d, s in tensor_sizes.items():
                if d not in self.axes:
                    out_sizes[d] = s

        self.dims, self.shape = zip(*out_sizes.items())
        self.digest = [crick.TDigest() for _ in range(int(numpy.prod(self.shape[1:])))]
        self.indices = product(*map(range, self.shape[1:]))

    def update_with_sample(self, sample: Sample):
        tensor = sample[self.tensor_name]
        assert "_percentiles" not in tensor.dims
        if self.digest is None:
            self._initialize(tensor.sizes)
            assert self.digest is not None

        for i, idx in enumerate(self.indices):
            self.digest[i].update(tensor.isel(dict(zip(self.dims[1:], idx))))

    def finalize(self) -> Dict[TensorName, Dict[Measure, MeasureValue]]:
        if self.digest is None:
            return {}
        else:
            vs = numpy.asarray([[d.quantile(q) for d in self.digest] for q in self.qs]).reshape(self.shape)
            return {
                self.tensor_name: {
                    Percentile(n=n, axes=self.axes): xr.DataArray(v, dims=self.dims[1:]) for n, v in zip(self.ns, vs)
                }
            }


if crick is None:
    DatasetPercentileGroup: Union[Type[MeanPercentiles], Type[CrickPercentiles]] = MeanPercentiles
else:
    DatasetPercentileGroup = CrickPercentiles


class SingleMeasureAsGroup(SampleMeasureGroup):
    """wrapper for measures to match interface of SampleMeasureGroup"""

    def __init__(self, tensor_name: TensorName, measure: Measure):
        self.tensor_name = tensor_name
        self.measure = measure

    def compute(self, sample: Sample) -> Dict[TensorName, Dict[Measure, MeasureValue]]:
        return {self.tensor_name: {self.measure: self.measure.compute(sample[self.tensor_name])}}


def get_measure_groups(measures: RequiredMeasures) -> MeasureGroups:
    """find a list of MeasureGroups to compute measures efficiently"""

    measure_groups = {PER_SAMPLE: [], PER_DATASET: []}
    means: Set[Tuple[TensorName, Mean]] = set()
    mean_var_std_groups: Set[Tuple[TensorName, Optional[Tuple[str, ...]]]] = set()
    percentile_groups: DefaultDict[Tuple[TensorName, Optional[Tuple[str, ...]]], List[float]] = defaultdict(list)
    for mode, ms_per_mode in measures.items():
        for tn, ms_per_tn in ms_per_mode.items():
            for m in ms_per_tn:
                if isinstance(m, Mean):
                    means.add((tn, m))
                elif isinstance(m, (Var, Std)):
                    mean_var_std_groups.add((tn, m.axes))
                elif isinstance(m, Percentile):
                    percentile_groups[(tn, m.axes)].append(m.n)
                elif mode == PER_SAMPLE:
                    measure_groups[mode].append(SingleMeasureAsGroup(tensor_name=tn, measure=m))
                else:
                    raise NotImplementedError(f"Computing statistics for {m} {mode} not yet implemented")

        # add all mean measures that are not included in a mean/var/std group
        for (tn, m) in means:
            if (tn, m.axes) not in mean_var_std_groups:
                # compute only mean
                if mode == PER_SAMPLE:
                    measure_groups[mode].append(SingleMeasureAsGroup(tensor_name=tn, measure=m))
                elif mode == PER_DATASET:
                    measure_groups[mode].append(DatasetMean(tensor_name=tn, axes=m.axes))
                else:
                    raise NotImplementedError(mode)

        for (tn, axes) in mean_var_std_groups:
            measure_groups[mode].append(MeanVarStd(tensor_name=tn, axes=axes))

        for (tn, axes), ns in percentile_groups.items():
            if mode == PER_SAMPLE:
                measure_groups[mode].append(SamplePercentiles(tensor_name=tn, axes=axes, ns=ns))
            elif mode == PER_DATASET:
                measure_groups[mode].append(DatasetPercentileGroup(tensor_name=tn, axes=axes, ns=ns))
            else:
                raise NotImplementedError(mode)

    return measure_groups


def compute_measures(
    measures: RequiredMeasures, *, sample: Optional[Sample] = None, dataset: Iterator[Sample] = tuple()
) -> ComputedMeasures:
    ms_groups = get_measure_groups(measures)
    ret = {PER_SAMPLE: {}, PER_DATASET: {}}
    if sample is not None:
        for mg in ms_groups[PER_SAMPLE]:
            assert isinstance(mg, SampleMeasureGroup)
            ret[PER_SAMPLE].update(mg.compute(sample))

    for sample in dataset:
        for mg in ms_groups[PER_DATASET]:
            assert isinstance(mg, DatasetMeasureGroup)
            mg.update_with_sample(sample)

    for mg in ms_groups[PER_DATASET]:
        assert isinstance(mg, DatasetMeasureGroup)
        ret[PER_DATASET].update(mg.finalize())

    return ret



import dataclasses
from typing import Any, Dict, List, Optional, Sequence, Union

from bioimageio.core.resource_io import nodes
from ._processing import AssertDtype, EnsureDtype, KNOWN_PROCESSING, Processing, TensorName
from ._utils import ComputedMeasures, PER_DATASET, PER_SAMPLE, RequiredMeasures, Sample

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore


@dataclasses.dataclass
class ProcessingInfoStep:
    name: str
    kwargs: Dict[str, Any]


@dataclasses.dataclass
class ProcessingInfo:
    steps: List[ProcessingInfoStep]
    assert_dtype_before: Optional[Union[str, Sequence[str]]] = None  # throw AssertionError if data type doesn't match
    ensure_dtype_before: Optional[str] = None  # cast data type if needed
    assert_dtype_after: Optional[Union[str, Sequence[str]]] = None  # throw AssertionError if data type doesn't match
    ensure_dtype_after: Optional[str] = None  # throw AssertionError if data type doesn't match


class CombinedProcessing:
    def __init__(self, combine_tensors: Dict[TensorName, ProcessingInfo]):
        self._procs = []
        known = dict(KNOWN_PROCESSING["pre"])
        known.update(KNOWN_PROCESSING["post"])

        # ensure all tensors have correct data type before any processing
        for tensor_name, info in combine_tensors.items():
            if info.assert_dtype_before is not None:
                self._procs.append(AssertDtype(tensor_name=tensor_name, dtype=info.assert_dtype_before))

            if info.ensure_dtype_before is not None:
                self._procs.append(EnsureDtype(tensor_name=tensor_name, dtype=info.ensure_dtype_before))

        for tensor_name, info in combine_tensors.items():
            for step in info.steps:
                self._procs.append(known[step.name](tensor_name=tensor_name, **step.kwargs))

            if info.assert_dtype_after is not None:
                self._procs.append(AssertDtype(tensor_name=tensor_name, dtype=info.assert_dtype_after))

            # ensure tensor has correct data type right after its processing
            if info.ensure_dtype_after is not None:
                self._procs.append(EnsureDtype(tensor_name=tensor_name, dtype=info.ensure_dtype_after))

        self.required_measures: RequiredMeasures = self._collect_required_measures(self._procs)
        self.tensor_names = list(combine_tensors)

    @classmethod
    def from_tensor_specs(cls, tensor_specs: List[Union[nodes.InputTensor, nodes.OutputTensor]]):
        combine_tensors = {}
        for ts in tensor_specs:
            # There is a difference between pre-and postprocessing:
            # After preprocessing we ensure float32, because the output is consumed by the model.
            # After postprocessing the dtype that is specified in the model spec needs to be ensured.
            assert ts.name not in combine_tensors
            if isinstance(ts, nodes.InputTensor):
                # todo: assert nodes.InputTensor.dtype with assert_dtype_before?
                # todo: in the long run we do not want to limit model inputs to float32...
                combine_tensors[ts.name] = ProcessingInfo(
                    [ProcessingInfoStep(p.name, kwargs=p.kwargs) for p in ts.preprocessing or []],
                    ensure_dtype_after="float32",
                )
            elif isinstance(ts, nodes.OutputTensor):
                combine_tensors[ts.name] = ProcessingInfo(
                    [ProcessingInfoStep(p.name, kwargs=p.kwargs) for p in ts.postprocessing or []],
                    ensure_dtype_after=ts.data_type,
                )
            else:
                raise NotImplementedError(type(ts))

        inst = cls(combine_tensors)
        for ts in tensor_specs:
            if isinstance(ts, nodes.OutputTensor) and ts.name in inst.required_measures[PER_DATASET]:
                raise NotImplementedError("computing statistics for output tensors per dataset is not yet implemented")

        return inst

    def apply(self, sample: Sample, computed_measures: ComputedMeasures) -> None:
        for proc in self._procs:
            proc.set_computed_measures(computed_measures)
            sample[proc.tensor_name] = proc.apply(sample[proc.tensor_name])

    @staticmethod
    def _collect_required_measures(proc: Sequence[Processing]) -> RequiredMeasures:
        ret: RequiredMeasures = {PER_SAMPLE: {}, PER_DATASET: {}}
        for p in proc:
            for mode, ms_per_mode in p.get_required_measures().items():
                for tn, ms_per_tn in ms_per_mode.items():
                    if tn not in ret[mode]:
                        ret[mode][tn] = set()

                    ret[mode][tn].update(ms_per_tn)

        return ret



from pathlib import Path

import imageio
import numpy as np
from numpy.testing import assert_array_almost_equal

from bioimageio.core import load_resource_description
from bioimageio.core.resource_io.nodes import Model


def test_predict_image(any_model, tmpdir):
    from bioimageio.core.prediction import predict_image

    spec = load_resource_description(any_model)
    assert isinstance(spec, Model)
    inputs = spec.test_inputs

    outputs = [Path(tmpdir) / f"out{i}.npy" for i in range(len(spec.test_outputs))]
    predict_image(any_model, inputs, outputs)
    for out_path in outputs:
        assert out_path.exists()

    result = [np.load(str(p)) for p in outputs]
    expected = [np.load(str(p)) for p in spec.test_outputs]
    for res, exp in zip(result, expected):
        assert_array_almost_equal(res, exp, decimal=4)


def test_predict_image_with_weight_format(unet2d_fixed_shape_or_not, tmpdir):
    from bioimageio.core.prediction import predict_image

    spec = load_resource_description(unet2d_fixed_shape_or_not)
    assert isinstance(spec, Model)
    inputs = spec.test_inputs

    outputs = [Path(tmpdir) / f"out{i}.npy" for i in range(len(spec.test_outputs))]
    predict_image(unet2d_fixed_shape_or_not, inputs, outputs, weight_format="pytorch_state_dict")
    for out_path in outputs:
        assert out_path.exists()

    result = [np.load(str(p)) for p in outputs]
    expected = [np.load(str(p)) for p in spec.test_outputs]
    for res, exp in zip(result, expected):
        assert_array_almost_equal(res, exp, decimal=4)


def _test_predict_with_padding(model, tmp_path):
    from bioimageio.core.prediction import predict_image

    spec = load_resource_description(model)
    assert isinstance(spec, Model)

    input_spec, output_spec = spec.inputs[0], spec.outputs[0]
    channel_axis = input_spec.axes.index("c")
    channel_first = channel_axis == 1

    image = np.load(str(spec.test_inputs[0]))
    assert image.shape[channel_axis] == 1
    if channel_first:
        image = image[0, 0]
    else:
        image = image[0, ..., 0]
    original_shape = image.shape
    assert image.ndim == 2

    if isinstance(output_spec.shape, list):
        n_channels = output_spec.shape[channel_axis]
    else:
        scale = output_spec.shape.scale[channel_axis]
        offset = output_spec.shape.offset[channel_axis]
        in_channels = 1
        n_channels = int(2 * offset + scale * in_channels)

    # write the padded image
    image = image[3:-2, 1:-12]
    in_path = tmp_path / "in.tif"
    out_path = tmp_path / "out.tif"
    imageio.imwrite(in_path, image)

    if hasattr(output_spec.shape, "scale"):
        scale = dict(zip(output_spec.axes, output_spec.shape.scale))
        offset = dict(zip(output_spec.axes, output_spec.shape.offset))
        spatial_axes = [ax for ax in output_spec.axes if ax in "xyz"]
        network_resizes = any(sc != 1 for ax, sc in scale.items() if ax in spatial_axes) or any(
            off != 0 for ax, off in offset.items() if ax in spatial_axes
        )
    else:
        network_resizes = False

    if network_resizes:
        exp_shape = tuple(int(sh * scale[ax] + 2 * offset[ax]) for sh, ax in zip(image.shape, spatial_axes))
    else:
        exp_shape = image.shape

    def check_result():
        if n_channels == 1:
            assert out_path.exists()
            res = imageio.imread(out_path)
            assert res.shape == exp_shape
        else:
            path = str(out_path)
            for c in range(n_channels):
                channel_out_path = Path(path.replace(".tif", f"-c{c}.tif"))
                assert channel_out_path.exists()
                res = imageio.imread(channel_out_path)
                assert res.shape == exp_shape

    # test with dynamic padding
    predict_image(model, in_path, out_path, padding={"x": 16, "y": 16, "mode": "dynamic"})
    check_result()

    # test with fixed padding
    predict_image(model, in_path, out_path, padding={"x": original_shape[0], "y": original_shape[1], "mode": "fixed"})
    check_result()

    # test with automated padding
    predict_image(model, in_path, out_path, padding=True)
    check_result()


# prediction with padding with the parameters above may not be suited for any model
# so we only run it for the pytorch unet2d here
def test_predict_image_with_padding(unet2d_fixed_shape_or_not, tmp_path):
    _test_predict_with_padding(unet2d_fixed_shape_or_not, tmp_path)


# and with different output shape
def test_predict_image_with_padding_diff_output_shape(unet2d_diff_output_shape, tmp_path):
    _test_predict_with_padding(unet2d_diff_output_shape, tmp_path)


def test_predict_image_with_padding_channel_last(stardist, tmp_path):
    _test_predict_with_padding(stardist, tmp_path)


def _test_predict_image_with_tiling(model, tmp_path: Path, exp_mean_deviation):
    from bioimageio.core.prediction import predict_image

    spec = load_resource_description(model)
    assert isinstance(spec, Model)
    inputs = spec.test_inputs
    assert len(inputs) == 1
    exp = np.load(str(spec.test_outputs[0]))

    out_path = tmp_path.with_suffix(".npy")

    def check_result():
        assert out_path.exists()
        res = np.load(out_path)
        assert res.shape == exp.shape
        # check that the mean deviation is smaller than the expected value
        # note that we can't use array_almost_equal here, because the numerical differences
        # between tiled and normal prediction are too large
        mean_deviation = np.abs(res - exp).mean()
        assert mean_deviation <= exp_mean_deviation

    # with tiling config
    tiling = {"halo": {"x": 32, "y": 32}, "tile": {"x": 256, "y": 256}}
    predict_image(model, inputs, [out_path], tiling=tiling)
    check_result()

    # with tiling determined from spec
    predict_image(model, inputs, [out_path], tiling=True)
    check_result()


# prediction with tiling with the parameters above may not be suited for any model
# so we only run it for the pytorch unet2d here
def test_predict_image_with_tiling_1(unet2d_nuclei_broad_model, tmp_path: Path):
    _test_predict_image_with_tiling(unet2d_nuclei_broad_model, tmp_path, 0.012)


def test_predict_image_with_tiling_2(unet2d_diff_output_shape, tmp_path: Path):
    _test_predict_image_with_tiling(unet2d_diff_output_shape, tmp_path, 0.06)


def test_predict_image_with_tiling_3(shape_change_model, tmp_path: Path):
    _test_predict_image_with_tiling(shape_change_model, tmp_path, 0.012)


def test_predict_image_with_tiling_channel_last(stardist, tmp_path: Path):
    _test_predict_image_with_tiling(stardist, tmp_path, 0.13)


def test_predict_image_with_tiling_fixed_output_shape(unet2d_fixed_shape, tmp_path: Path):
    _test_predict_image_with_tiling(unet2d_fixed_shape, tmp_path, 0.025)


def test_predict_images(unet2d_nuclei_broad_model, tmp_path: Path):
    from bioimageio.core.prediction import predict_images

    n_images = 5
    shape = (256, 256)

    in_paths = []
    out_paths = []
    for i in range(n_images):
        in_path = tmp_path / f"in{i}.tif"
        im = np.random.randint(0, 255, size=shape).astype("uint8")
        imageio.imwrite(in_path, im)
        in_paths.append(in_path)
        out_paths.append(tmp_path / f"out{i}.tif")
    predict_images(unet2d_nuclei_broad_model, in_paths, out_paths)

    for outp in out_paths:
        assert outp.exists()
        out = imageio.imread(outp)
        assert out.shape == shape



import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_array_almost_equal

from bioimageio.core import load_resource_description
from bioimageio.core.resource_io.nodes import Model
from bioimageio.core.utils import skip_on


class TooFewDevicesException(Exception):
    pass


def _test_device_management(model_package, weight_format):
    import torch

    if torch.cuda.device_count() == 0:
        raise TooFewDevicesException("Need at least one cuda device for this test")

    from bioimageio.core.prediction_pipeline import create_prediction_pipeline

    bio_model = load_resource_description(model_package)
    assert isinstance(bio_model, Model)
    pred_pipe = create_prediction_pipeline(bioimageio_model=bio_model, weight_format=weight_format, devices=["cuda:0"])

    inputs = [
        xr.DataArray(np.load(str(test_tensor)), dims=tuple(spec.axes))
        for test_tensor, spec in zip(bio_model.test_inputs, bio_model.inputs)
    ]
    with pred_pipe as pp:
        outputs = pp.forward(*inputs)

    assert isinstance(outputs, list)

    expected_outputs = [
        xr.DataArray(np.load(str(test_tensor)), dims=tuple(spec.axes))
        for test_tensor, spec in zip(bio_model.test_outputs, bio_model.outputs)
    ]

    assert len(outputs) == len(expected_outputs)
    for out, exp in zip(outputs, expected_outputs):
        assert_array_almost_equal(out, exp, decimal=4)

    # repeat inference with context manager to test load/unload/load/forward
    with pred_pipe as pp:
        outputs = pp.forward(*inputs)

    assert len(outputs) == len(expected_outputs)
    for out, exp in zip(outputs, expected_outputs):
        assert_array_almost_equal(out, exp, decimal=4)


@skip_on(TooFewDevicesException, reason="Too few devices")
def test_device_management_torch(any_torch_model):
    _test_device_management(any_torch_model, "pytorch_state_dict")


@skip_on(TooFewDevicesException, reason="Too few devices")
def test_device_management_torchscript(any_torchscript_model):
    _test_device_management(any_torchscript_model, "torchscript")


@pytest.mark.skipif(pytest.skip_torch, reason="requires torch for device discovery")
@skip_on(TooFewDevicesException, reason="Too few devices")
def test_device_management_onnx(any_onnx_model):
    _test_device_management(any_onnx_model, "onnx")


@pytest.mark.skipif(pytest.skip_torch, reason="requires torch for device discovery")
@skip_on(TooFewDevicesException, reason="Too few devices")
def test_device_management_tensorflow(any_tensorflow_model):
    _test_device_management(any_tensorflow_model, "tensorflow_saved_model_bundle")


@pytest.mark.skipif(pytest.skip_torch, reason="requires torch for device discovery")
@skip_on(TooFewDevicesException, reason="Too few devices")
def test_device_management_keras(any_keras_model):
    _test_device_management(any_keras_model, "keras_hdf5")



import json
import pathlib

__version__ = json.loads((pathlib.Path(__file__).parent / "VERSION").read_text())["version"]

from .resource_io import (
    export_resource_package,
    load_raw_resource_description,
    load_resource_description,
    save_raw_resource_description,
    serialize_raw_resource_description,
)
from .prediction_pipeline import create_prediction_pipeline
from .prediction import predict_image, predict_images, predict_with_padding, predict_with_tiling
from .resource_tests import check_input_shape, check_output_shape, test_resource



import json
from pathlib import Path

from setuptools import find_namespace_packages, setup

# Get the long description from the README file
ROOT_DIR = Path(__file__).parent.resolve()
long_description = (ROOT_DIR / "README.md").read_text(encoding="utf-8")
VERSION_FILE = ROOT_DIR / "bioimageio" / "core" / "VERSION"
VERSION = json.loads(VERSION_FILE.read_text())["version"]


setup(
    name="bioimageio.core",
    version=VERSION,
    description="Python functionality for the bioimage model zoo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bioimage-io/core-bioimage-io-python",
    author="Bioimage Team",
    classifiers=[  # Optional
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_namespace_packages(exclude=["tests"]),  # Required
    install_requires=[
        "bioimageio.spec==0.4.9",
        "imageio>=2.5",
        "numpy",
        "ruamel.yaml",
        "tqdm",
        "xarray",
        "tifffile",
    ],
    include_package_data=True,
    extras_require={
        "test": ["pytest", "black", "mypy"],
        "dev": ["pre-commit"],
        "pytorch": ["torch>=1.6", "torchvision"],
        "tensorflow": ["tensorflow"],
        "onnx": ["onnxruntime"],
    },
    project_urls={  # Optional
        "Bug Reports": "https://github.com/bioimage-io/core-bioimage-io-python/issues",
        "Source": "https://github.com/bioimage-io/core-bioimage-io-python",
    },
    entry_points={"console_scripts": ["bioimageio = bioimageio.core.__main__:app"]},
)



import shutil
import traceback
from pathlib import Path
from typing import List, Optional, Union

from bioimageio.core import export_resource_package
from bioimageio.core.resource_io.utils import resolve_source
from bioimageio.spec.commands import validate
from bioimageio.spec.shared.raw_nodes import URI


def package(
    rdf_source: Union[Path, str, URI, dict],
    path: Path = Path() / "{src_name}-package.zip",
    weights_priority_order: Optional[List[str]] = None,
    verbose: bool = False,
) -> int:
    """Package a BioImage.IO resource described by a BioImage.IO Resource Description File (RDF)."""
    code = validate(rdf_source, update_format=True, update_format_inner=True)
    source_name = rdf_source.get("name") if isinstance(rdf_source, dict) else rdf_source
    if code["error"]:
        print(f"Cannot package invalid BioImage.IO RDF {source_name}")
        return 1

    try:
        tmp_package_path = export_resource_package(rdf_source, weights_priority_order=weights_priority_order)
    except Exception as e:
        print(f"Failed to package {source_name} due to: {e}")
        if verbose:
            traceback.print_exc()
        return 1

    try:
        rdf_local_source = resolve_source(rdf_source)
    except Exception as e:
        print(f"Failed to resolve RDF source {rdf_source}: {e}")
        if verbose:
            traceback.print_exc()
        return 1

    try:
        path = path.with_name(path.name.format(src_name=rdf_local_source.stem))
        shutil.move(tmp_package_path, path)
    except Exception as e:
        print(f"Failed to move package from {tmp_package_path} to {path} due to: {e}")
        if verbose:
            traceback.print_exc()
        return 1

    print(f"exported bioimageio package from {source_name} to {path}")
    return 0



import json
import subprocess
import sys

import pytest
from packaging.version import Version


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python 3.8")
@pytest.mark.skipif(pytest.mamba_cmd is None, reason="requires mamba")
def test_bioimageio_spec_version():
    from importlib.metadata import metadata

    # get latest released bioimageio.spec version
    mamba_repoquery = subprocess.run(
        f"{pytest.mamba_cmd} repoquery search -c conda-forge --json bioimageio.spec".split(" "),
        encoding="utf-8",
        capture_output=True,
        check=True,
    )
    full_out = mamba_repoquery.stdout  # full output includes mamba banner
    search = json.loads(full_out[full_out.find("{") :])  # json output starts at '{'
    rmaj, rmin, rpatch, *_ = search["result"]["pkgs"][0]["version"].split(".")
    released = Version(f"{rmaj}.{rmin}.{rpatch}")

    # get currently pinned bioimageio.spec version
    meta = metadata("bioimageio.core")
    req = meta["Requires-Dist"]
    print(req)
    assert req.startswith("bioimageio.spec ==")
    spec_ver = req[len("bioimageio.spec ==") :]
    assert spec_ver.count(".") == 2
    pmaj, pmin, ppatchand_and_post = spec_ver.split(".")
    assert (ppatchand_and_post.isdigit() or ppatchand_and_post[:-1].isdigit()) and (
        ppatchand_and_post[-1] == "*" or ppatchand_and_post[-1].isdigit()
    ), "bioimageio.spec version should be pinned down to patch, e.g. '0.4.9*'"

    ppatch = ppatchand_and_post[:-1] if ppatchand_and_post[-1] == "*" else ppatchand_and_post
    pinned = Version(f"{pmaj}.{pmin}.{ppatch}")

    assert pinned >= released, "bioimageio.spec pinned to an old version!"
